﻿Imports System.Math
Imports System.Threading


Partial Public Class classArm

    ' angles in degrees
    Public Structure typeArmAngles
        Dim sngBaseYX As Single
        Dim sngBaseXZ As Single
        Dim sngShoulder As Single
        Dim sngElbow As Single
        Dim sngWrist As Single
        Dim sngWristRotate As Single
        Dim sngGripperOpening As Single  ' not an angle.  approx opening in mm
    End Structure

    Public taaArmAngles As typeArmAngles  '  NOT the joint angles
    Public tjcJointCoordinates As typeJointCoordinate


    Public Const constGripperOpenPosition = 519
    Public Const constGripperClosedPosition = 718

    Public Const constDefaultPortNum As Byte = 3  'COM3
    Public Const constDefaultBaudNum As Byte = 1  '1Mbps

    ' Dynamixel Register Addresses (see http://support.robotis.com/en/product/dynamixel/ax_series/dxl_ax_actuator.htm)
    Const constPresentPosition As Short = 36
    Public Const constGoalPosition As Short = 30
    Public Const constCWComplianceMargin As Short = 26
    Public Const constCCWComplianceMargin As Short = 27
    Public Const constMovingSpeed As Short = 32
    Public Const constAlarmShutDown As Short = 18
    Public Const constTorqueLimit As Short = 34
    Public Const constTorqueMax As Short = 14
    Public Const constAlarm As Short = 17
    Public Const constLoad As Short = 40
    Public Const constTemperature As Short = 43

    Public constShoulderSegment As Byte = 2
    Public constElbowSegment As Byte = 3
    Public constWristRSegment As Byte = 4
    Public constWristSegment As Byte = 5
    Public constGripperSegment As Byte = 6



    'Coordinates in mm
    Public Structure typeJointCoordinate
        Dim tpcShoulderBase As typePointCoordinate
        Dim tpcShoulder As typePointCoordinate
        Dim tpcElbow As typePointCoordinate
        Dim tpcWrist As typePointCoordinate
        Dim tpcWristRotateB4Offset As typePointCoordinate
        Dim tpcWristRotate As typePointCoordinate
        Dim tpcGripperEnd As typePointCoordinate
    End Structure


    'Misc
    Public boolArmSimulationMode As Boolean = False
    Public intArmSegmentSelect As Short = 2


    Public Sub frmGotAnArm_Load(sender As Object, e As EventArgs) 'Handles MyBase.Load
        Try
            Me.Visible = True
            dxl_initialize(constDefaultPortNum, constDefaultBaudNum)
            subInitializeDynamixels()
            tsslJoint.Text = "Shoulder"
            subHome(True)
            taaArmAngles = ftnForwardKinematicsAngles()
            tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            subDrawArm()
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION In frmGotAnArm_Load " & ex.ToString)
        End Try
    End Sub

    Public Sub subArmExit() Handles Me.FormClosing
        dxl_terminate()
    End Sub

    ' =================================   MANUAL CONTROL ========================================
    Private Sub subSegment_MouseWheel(ByVal sender As System.Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseWheel  '
        Try
            taaArmAngles = ftnForwardKinematicsAngles()
            tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            Select Case intArmSegmentSelect
                Case 0
                Case 1
                Case 2
                    If boolArmSimulationMode Then
                        taaArmAngles.sngShoulder += e.Delta / 250
                    Else
                        ftnSetShoulderPosition(ftnNewPosition(3, e.Delta))
                    End If
                Case 3
                    If boolArmSimulationMode Then
                        taaArmAngles.sngElbow += e.Delta / 250
                    Else
                        ftnSetElbowPosition(ftnNewPosition(4, -e.Delta))
                    End If
                Case 4
                    If boolArmSimulationMode Then
                        taaArmAngles.sngWristRotate += e.Delta / 250
                    Else
                        ftnSetWristRotatePosition(ftnNewPosition(6, e.Delta))
                    End If
                Case 5
                Case 6
                    If boolArmSimulationMode Then
                        taaArmAngles.sngGripperOpening += e.Delta / 200
                    Else
                        ftnSetGripperPosition(ftnNewPosition(7, e.Delta))
                    End If
            End Select
            tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            subDrawArm()
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION In subSegment_MouseWheel " & ex.ToString)
        End Try
    End Sub

    Function ftnNewPosition(ByVal intID As Short, ByVal intIncrement As Short) As Integer
        Try
            ftnNewPosition = dxl_read_word(intID, constPresentPosition) + intIncrement / 10
            If (ftnNewPosition > 1023) Then ftnNewPosition = 1023
            If (ftnNewPosition < 0) Then ftnNewPosition = 0
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION In ftnNewPosition " & ex.ToString)
        End Try
    End Function

    Dim boolMouseDown As Boolean = False
    Private Sub subArm_MouseDown(ByVal sender As System.Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles pbbArm.MouseDown
        Try
            boolMouseDown = True
            If (e.Button = System.Windows.Forms.MouseButtons.Right) Then  ' Select the next arm segment
                intArmSegmentSelect += 1
                If (intArmSegmentSelect > 6) Then intArmSegmentSelect = 2
                Select Case intArmSegmentSelect
                    Case 0
                        tsslJoint.Text = "---"
                    Case 1
                        tsslJoint.Text = "---"
                    Case constShoulderSegment
                        tsslJoint.Text = "Shoulder"
                    Case constElbowSegment
                        tsslJoint.Text = "Elbow"
                    Case constWristRSegment
                        tsslJoint.Text = "Wrist Rotate"
                    Case constWristSegment
                        tsslJoint.Text = "Wrist"
                    Case constGripperSegment
                        tsslJoint.Text = "Gripper"
                    Case Else
                        tsslJoint.Text = "-----"
                End Select
            ElseIf (e.Button = System.Windows.Forms.MouseButtons.Left) Then  ' Set a destination
                Dim tmcDestination As typeMapCoordinate = ftnMousePosition()
                Dim tpcDestination As typePointCoordinate
                tpcDestination.X = tmcDestination.X
                tpcDestination.Z = tmcDestination.Y
                If booldebugArm Then frmDebug.subWriteCoordinate(" subArm_MouseDown: tmcDestination ", tmcDestination)
                Dim taaArmAnglesDesired As typeArmAngles = taaArmAngles
                Dim tjcJointCoordinatesDesired As typeJointCoordinate = tjcJointCoordinates
                Dim boolFoundSolution As Boolean = ftnReverseKinematics(tpcDestination, taaArmAnglesDesired, tjcJointCoordinatesDesired)
                If booldebugArm Then frmDebug.subWriteLogs(" subArm_MouseDown: boolFoundSolution = " & boolFoundSolution)
                If boolFoundSolution Then
                    ftnSetShoulderPosition(ftnDegreeToPosition(taaArmAnglesDesired.sngShoulder - taaArmAnglesDesired.sngBaseXZ))
                    ftnSetElbowPosition(ftnDegreeToPosition(ftnConvertTo180(taaArmAnglesDesired.sngElbow - taaArmAnglesDesired.sngShoulder)))
                End If
            ElseIf (e.Button = System.Windows.Forms.MouseButtons.Middle) Then  ' Set gripper orientation
                Dim tmcPoint As typeMapCoordinate = ftnMousePosition()
                Dim sngOrientation As Single = (tmcPoint.Y) / pbbArm.Height * 180
                Dim taaArmAnglesDesired As typeArmAngles = taaArmAngles
                Dim tjcJointCoordinatesDesired As typeJointCoordinate = tjcJointCoordinates
                Dim boolFoundSolution As Boolean = ftnReverseKinematics(sngOrientation, taaArmAnglesDesired, tjcJointCoordinatesDesired)
                If boolFoundSolution Then
                    ftnSetShoulderPosition(ftnDegreeToPosition(taaArmAnglesDesired.sngShoulder))
                    ftnSetElbowPosition(ftnDegreeToPosition(ftnConvertTo180(taaArmAnglesDesired.sngElbow - taaArmAnglesDesired.sngShoulder)))
                End If
            End If
            subDrawArm()
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in subArm_MouseDown " & ex.ToString)
        End Try
    End Sub

    Private Sub subArm_MouseUp(ByVal sender As System.Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles pbbArm.MouseUp
        On Error Resume Next
        boolMouseDown = False
    End Sub


    Private Sub subArm_MouseMove(ByVal sender As System.Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles pbbArm.MouseMove  'base, shoulder, elbow, wrist control
        On Error Resume Next
        timerUpdate.Enabled = False
        Dim tmcDestination As typeMapCoordinate = ftnMousePosition()
        tsslXY.Text = tmcDestination.X & ", " & tmcDestination.Y
        timerUpdate.Enabled = True
    End Sub

    Function ftnMousePosition() As typeMapCoordinate
        Try
            ftnMousePosition.X = -(pbbArm.Width * 0.25 - (MousePosition.X - Me.Location.X - pbbArm.Location.X - pbbArm.Padding.Left - pbbArm.Margin.Left - 3))
            ftnMousePosition.Y = (pbbArm.Height * 0.5 - (MousePosition.Y - Me.Location.Y - pbbArm.Location.Y - pbbArm.Padding.Top - pbbArm.Margin.Top - 28))
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION In ftnMousePosition " & ex.ToString)
        End Try
    End Function

    ' ============================ Higher Level Arm Control  ==========================================

    Public Function ftnGetGripperEndPoint() As typePointCoordinate
        Try
            ftnGetGripperEndPoint = tjcJointCoordinates.tpcGripperEnd
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnGetGripperEndPoint " & ex.ToString)
        End Try
    End Function

    Public Function ftnGetCameraPose() As typeBotPose  ' NOTE: Orientation is in xz plane.  x and y are in mm
        Try
            taaArmAngles = ftnForwardKinematicsAngles()
            tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            ftnGetCameraPose.Location.X = tjcJointCoordinates.tpcWrist.X
            ftnGetCameraPose.Location.Y = tjcJointCoordinates.tpcWrist.Z
            ftnGetCameraPose.Orientation = taaArmAngles.sngWrist
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnGetCameraPose " & ex.ToString)
        End Try
    End Function

    Public Function ftnGetShoulderBaseCoordinate() As typeMapCoordinate
        Try
            ftnGetShoulderBaseCoordinate.X = 3
            ftnGetShoulderBaseCoordinate.Y = 0
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnGetCameraPose " & ex.ToString)
        End Try
    End Function

    Public Function ftnSetArmEndPoint(ByRef tpcDestination As typePointCoordinate) As Boolean
        Try
            ftnSetArmEndPoint = False
            Dim taaArmAnglesDesired As typeArmAngles = taaArmAngles
            Dim tjcJointCoordinatesDesired As typeJointCoordinate = tjcJointCoordinates
            ftnSetArmEndPoint = ftnReverseKinematics(tpcDestination, taaArmAnglesDesired, tjcJointCoordinatesDesired)
            frmDebug.subWriteLogs(" subArm_MouseDown: ftnSetArmEndPoint = " & ftnSetArmEndPoint)
            If ftnSetArmEndPoint Then
                ftnSetShoulderPosition(ftnDegreeToPosition(taaArmAnglesDesired.sngShoulder - taaArmAnglesDesired.sngBaseXZ))
                ftnSetElbowPosition(ftnDegreeToPosition(ftnConvertTo180(taaArmAnglesDesired.sngElbow - taaArmAnglesDesired.sngShoulder)))
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnSetArmEndPoint " & ex.ToString)
        End Try
    End Function

    Public Function ftnSetGripperOrientation(ByVal sngGripperOrientation As Single, ByVal boolHigh As Boolean, ByVal boolLow As Boolean, ByVal boolMid As Boolean) As Boolean
        Try
            ftnSetGripperOrientation = False
            Dim taaArmAnglesDesired As typeArmAngles = ftnForwardKinematicsAngles()
            Dim tjcJointCoordinatesDesired As typeJointCoordinate = ftnForwardKinematicsCoordinates(taaArmAngles)
            If booldebugArm Then
                frmDebug.subWriteLogs(" ---------------------------------------------------------------------------------------")
                frmDebug.subWriteLogs(" ftnSetGripperOrientation: original taaArmAnglesDesired.sngShoulder = " & taaArmAnglesDesired.sngShoulder)
                frmDebug.subWriteLogs(" ftnSetGripperOrientation: original taaArmAnglesDesired.sngElbow = " & taaArmAnglesDesired.sngElbow)
                frmDebug.subWriteLogs(" ftnSetGripperOrientation: original Elbow joint angle = " & ftnConvertTo180(taaArmAnglesDesired.sngElbow - taaArmAnglesDesired.sngShoulder))
            End If
            If boolHigh Then
                taaArmAnglesDesired.sngShoulder = 90
                taaArmAnglesDesired.sngElbow = 0
                taaArmAnglesDesired.sngWrist = taaArmAnglesDesired.sngElbow
                taaArmAnglesDesired.sngWristRotate = taaArmAnglesDesired.sngElbow
            ElseIf boolLow Then
                taaArmAnglesDesired.sngShoulder = -70
                taaArmAnglesDesired.sngElbow = 0
                taaArmAnglesDesired.sngWrist = taaArmAnglesDesired.sngElbow
                taaArmAnglesDesired.sngWristRotate = taaArmAnglesDesired.sngElbow
            ElseIf boolMid Then
                taaArmAnglesDesired.sngShoulder = 0
                taaArmAnglesDesired.sngElbow = 0
                taaArmAnglesDesired.sngWrist = taaArmAnglesDesired.sngElbow
                taaArmAnglesDesired.sngWristRotate = taaArmAnglesDesired.sngElbow
            End If
            Dim boolFoundSolution As Boolean = ftnReverseKinematics(sngGripperOrientation, taaArmAnglesDesired, tjcJointCoordinatesDesired)
            frmDebug.subWriteLogs(" ftnSetGripperOrientation: boolFoundSolution = " & boolFoundSolution)
            If boolFoundSolution Then
                If booldebugArm Then
                    frmDebug.subWriteLogs(" ftnSetGripperOrientation: taaArmAnglesDesired.sngShoulder = " & taaArmAnglesDesired.sngShoulder)
                    frmDebug.subWriteLogs(" ftnSetGripperOrientation: taaArmAnglesDesired.sngElbow = " & taaArmAnglesDesired.sngElbow)
                    frmDebug.subWriteLogs(" ftnSetGripperOrientation: Elbow joint angle = " & ftnConvertTo180(taaArmAnglesDesired.sngElbow - taaArmAnglesDesired.sngShoulder))
                End If
                ftnSetShoulderPosition(ftnDegreeToPosition(taaArmAnglesDesired.sngShoulder))
                ftnSetElbowPosition(ftnDegreeToPosition(ftnConvertTo180(taaArmAnglesDesired.sngElbow - taaArmAnglesDesired.sngShoulder)))
                taaArmAngles = taaArmAnglesDesired
                subDrawArm()
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnSetGripperOrientation " & ex.ToString)
        End Try
    End Function


    Public Function ftnPointGripperAtPoint(ByRef tpcPoint As typePointCoordinate) As Boolean
        Try
            ftnPointGripperAtPoint = False
            Dim boolHigh, boolLow, boolMid As Boolean
            boolHigh = tpcPoint.Z > 90
            taaArmAngles.sngElbow = 0
            taaArmAngles.sngWrist = taaArmAngles.sngElbow
            taaArmAngles.sngWristRotate = taaArmAngles.sngElbow
            If (tpcPoint.Z > 90) Then
                boolHigh = True
                boolMid = False
                boolLow = False
                taaArmAngles.sngShoulder = 90
                taaArmAngles.sngElbow = ftnConvertTo180(ftnRadianToDegree(Atan((tpcPoint.Z - tjcJointCoordinates.tpcElbow.Z) / tpcPoint.X)))
                tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            ElseIf (tpcPoint.Z < -90) Then
                boolHigh = False
                boolMid = False
                boolLow = True
                taaArmAngles.sngShoulder = -90
                taaArmAngles.sngElbow = ftnConvertTo180(ftnRadianToDegree(Atan((tpcPoint.Z - tjcJointCoordinates.tpcElbow.Z) / tpcPoint.X)))
                tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            Else
                boolHigh = False
                boolMid = False
                boolLow = False
                taaArmAngles.sngShoulder = 0
                taaArmAngles.sngElbow = ftnConvertTo180(ftnRadianToDegree(Atan((tpcPoint.Z - tjcJointCoordinates.tpcElbow.Z) / tpcPoint.X)))
                tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            End If
            ftnSetGripperOrientation(taaArmAngles.sngElbow, boolHigh, boolLow, boolMid)
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnPointGripperAtPoint " & ex.ToString)
        End Try
    End Function



    ' ============================Timers==========================================

    Private Sub timerWatchDog_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles timerWatchDog.Tick
        Try
            If Me.Visible Then
                timerWatchDog.Enabled = False
                Dim boolOK As Boolean = False
                For intID As Short = 2 To 7 ' check that USB2Dynamixel still working
                    dxl_ping(intID)
                    If (dxl_get_result() = COMM_RXSUCCESS) Then
                        boolOK = True
                        Exit For
                    End If
                Next
                If Not boolOK Then subInitializeDynamixels()
                If Not boolArmSimulationMode And Me.Visible Then
                    For intID As Short = 2 To 7 ' check for overload, overtemperature
                        dxl_ping(intID)
                        If (timeJointMove(intID).AddSeconds(12) < Now()) Then
                            If (dxl_read_word(intID, constLoad) > 1000) Then
                                dxl_write_word(intID, constGoalPosition, dxl_read_word(intID, constPresentPosition))
                                dxl_write_word(intID, constTorqueLimit, dxl_read_word(intID, constTorqueMax))
                                frmDebug.subWriteLogs(" timerWatchDog_Tick:  Overload intID =  " & intID)
                            End If
                        End If
                        If (dynamixel.dxl_get_rxpacket_error(dynamixel.ERRBIT_OVERLOAD) = 1) Or (dynamixel.dxl_get_rxpacket_error(dynamixel.ERRBIT_OVERHEAT) = 1) Then
                            dxl_write_word(intID, constGoalPosition, dxl_read_word(intID, constPresentPosition))
                            dxl_write_word(intID, constTorqueLimit, dxl_read_word(intID, constTorqueMax))
                            subHighlightDynamixelError(intID)
                        End If
                    Next
                End If
            End If
            timerWatchDog.Enabled = True
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in timerWatchDog_Tick " & ex.ToString)
        End Try
    End Sub

    Private Sub subTimerUpdate(sender As Object, e As EventArgs) Handles timerUpdate.Tick
        Try
            If Me.Visible And Not boolArmSimulationMode Then
                timerUpdate.Enabled = False
                taaArmAngles = ftnForwardKinematicsAngles()
                tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
                subDrawArm()
                tsslPosition.Text = tjcJointCoordinates.tpcGripperEnd.X & ",  " & tjcJointCoordinates.tpcGripperEnd.Y & ",  " & tjcJointCoordinates.tpcGripperEnd.Z
                subUpdateDynaParameters()
                tsslAngles.Text = taaArmAngles.sngShoulder.ToString & ", " & ftnGetJointAngle(4).ToString
                timerUpdate.Enabled = True
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in subTimerUpdate " & ex.ToString)
        End Try
    End Sub


    ' ======================================================================

    Sub subUpdateDynaParameters()
        Try
            Dim strText As String = "ID" & "  " & "Position" & vbTab & "Goal" & vbTab & "Load" & vbTab & "Temperature"
            If Not boolArmSimulationMode Then
                For intID As Short = 2 To 7
                    strText &= vbCrLf & intID.ToString & "       " & dxl_read_word(intID, constPresentPosition) & vbTab & dxl_read_word(intID, constGoalPosition) & vbTab _
                            & dxl_read_word(intID, constLoad) & vbTab & dxl_read_byte(intID, constTemperature)
                Next intID
            End If
            rtbDynaParam.Text = strText
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION In subUpdateDynaParameters " & ex.ToString)
        End Try
    End Sub


    ' =============================== BUTTONS =============================================================

    Public Sub subHome(ByVal boolGripperAdjust As Boolean)
        Try
            If boolArmSimulationMode Then
                taaArmAngles.sngShoulder = ftnPositionToDegree(920)
                taaArmAngles.sngElbow = 0
                taaArmAngles.sngWrist = 0   '
                taaArmAngles.sngWristRotate = 0
                If boolGripperAdjust Then taaArmAngles.sngGripperOpening = ftnPositionToDegree(465)
                tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            Else
                ftnSetShoulderPosition(920)
                ftnSetElbowPosition(120)
                ftnSetWristRotatePosition(512)
                If boolGripperAdjust Then ftnSetGripperPosition(465)
            End If
            subDrawArm()
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION In subHome " & ex.ToString)
        End Try
    End Sub

    Private Sub butReset_Click(sender As Object, e As EventArgs) Handles butReset.Click
        Try
            If boolArmSimulationMode Then
                subHome(False)
            Else
                butStop_Click(sender, e)
                For intID As Short = 2 To 7
                    If (dxl_read_byte(intID, constAlarmShutDown) <> 0) Then dxl_write_word(intID, constTorqueLimit, dxl_read_word(intID, constTorqueMax))
                Next
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION In butReset_Click " & ex.ToString)
        End Try
    End Sub

    Private Sub butHome_Click(sender As Object, e As EventArgs) Handles butHome.Click
        On Error Resume Next
        subHome(True)
    End Sub

    Private Sub butStop_Click(sender As Object, e As EventArgs) Handles butStop.Click
        Try
            If boolArmSimulationMode Then
            Else
                For intID As Short = 2 To 7
                    dxl_write_word(intID, constGoalPosition, dxl_read_word(intID, constPresentPosition))
                    dxl_write_word(intID, constTorqueLimit, dxl_read_word(intID, constTorqueMax))
                Next
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION In butStop_Click " & ex.ToString)
        End Try
    End Sub



End Class
