﻿
Imports System.Math

Partial Public Class classArm
    Const maxJointAngle = 125

    ' NOTE - dimensions are mmm and degrees

    'http://www.cerebralmeltdown.com/Robot%20Arm/Robot%20Arm%20Math/index.html

    Public Function ftnForwardKinematicsAngles() As typeArmAngles
        ' Arm dimensions in mm
        Try
            ftnForwardKinematicsAngles.sngShoulder = ftnGetJointAngle(3)
            ftnForwardKinematicsAngles.sngElbow = ftnConvertTo180(taaArmAngles.sngShoulder + ftnGetJointAngle(4))
            ftnForwardKinematicsAngles.sngWrist = taaArmAngles.sngElbow   '
            ftnForwardKinematicsAngles.sngWristRotate = taaArmAngles.sngElbow
            ftnForwardKinematicsAngles.sngGripperOpening = ftnGetJointAngle(7)
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnForwardKinematicsAngles " & ex.ToString)
        End Try
    End Function

    Public Function ftnForwardKinematicsCoordinates(ByRef taaArmAngles As typeArmAngles) As typeJointCoordinate
        ' Arm dimensions in mm
        Const constShoulderToElbow = 175   ' 
        Const constElbowToWristR = 67       ' 
        Const constWristROffset = 22
        Const constWristRToWrist = 0
        Const constWristToGripper = 108
        Const constBaseAngleXZ = 0  ' 
        Try
            ftnForwardKinematicsCoordinates.tpcShoulderBase.Z = 0
            ftnForwardKinematicsCoordinates.tpcShoulder = ftnPointFromPoint(ftnForwardKinematicsCoordinates.tpcShoulderBase, 0, constBaseAngleXZ)
            ftnForwardKinematicsCoordinates.tpcElbow = ftnPointFromPoint(ftnForwardKinematicsCoordinates.tpcShoulder, constShoulderToElbow, taaArmAngles.sngShoulder)
            ftnForwardKinematicsCoordinates.tpcWristRotateB4Offset = ftnPointFromPoint(ftnForwardKinematicsCoordinates.tpcElbow, constElbowToWristR, taaArmAngles.sngElbow)
            ftnForwardKinematicsCoordinates.tpcWristRotate = ftnPointFromPoint(ftnForwardKinematicsCoordinates.tpcWristRotateB4Offset, constWristROffset, ftnConvertTo180(taaArmAngles.sngWrist - 90))
            ftnForwardKinematicsCoordinates.tpcWrist = ftnPointFromPoint(ftnForwardKinematicsCoordinates.tpcWristRotate, constWristRToWrist, taaArmAngles.sngElbow)
            ftnForwardKinematicsCoordinates.tpcGripperEnd = ftnPointFromPoint(ftnForwardKinematicsCoordinates.tpcWrist, constWristToGripper, taaArmAngles.sngWrist)
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnForwardKinematicsCoordinates " & ex.ToString)
        End Try
    End Function


    '===============================================  Reverse Kinematics to get desired Gripper Orientation  ================================

    Public Function ftnReverseKinematics(ByVal sngGripperAngleDesired As Single, ByRef taaArmAngles As typeArmAngles, ByRef tjcJointCoordinates As typeJointCoordinate) As Boolean
        Try
            ftnReverseKinematics = False
            Dim sngStartingAngleShoulder As Single = taaArmAngles.sngShoulder
            Dim sngShoulderAngle As Single

            frmDebug.subWriteLogs(" ftnReverseKinematics: original taaArmAngles.sngShoulder = " & taaArmAngles.sngShoulder)
            frmDebug.subWriteLogs(" ftnReverseKinematics: original taaArmAngles.sngElbow = " & taaArmAngles.sngElbow)

            For sngShoulderAngle = sngStartingAngleShoulder To maxJointAngle Step 5
                ftnReverseKinematics = ftnValidDesiredKinematics(sngGripperAngleDesired, sngShoulderAngle) ' + shoulder 
                If ftnReverseKinematics Then Exit For
            Next sngShoulderAngle

            If Not ftnReverseKinematics Then
                For sngShoulderAngle = sngStartingAngleShoulder To -maxJointAngle Step -5
                    ftnReverseKinematics = ftnValidDesiredKinematics(sngGripperAngleDesired, sngShoulderAngle) ' + shoulder 
                    If ftnReverseKinematics Then Exit For
                Next sngShoulderAngle
            End If
            If Not ftnReverseKinematics Then
                Beep()
            Else
                taaArmAngles.sngShoulder = sngShoulderAngle
                taaArmAngles.sngWrist = sngGripperAngleDesired
                taaArmAngles.sngWristRotate = sngGripperAngleDesired
                taaArmAngles.sngElbow = ftnConvertTo180(taaArmAngles.sngShoulder + taaArmAngles.sngWrist)
                tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            End If
            frmDebug.subWriteLogs(" ftnReverseKinematics:  = " & ftnReverseKinematics)
            frmDebug.subWriteLogs(" ftnReverseKinematics: taaArmAngles.sngShoulder = " & taaArmAngles.sngShoulder)
            frmDebug.subWriteLogs(" ftnReverseKinematics: taaArmAngles.sngElbow = " & taaArmAngles.sngElbow)

        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnReverseKinematics " & ex.ToString)
        End Try
    End Function

    '===============================================  Reverse Kinematics to get desired Gripper Position  ================================

    ' http://freespace.virgin.net/hugo.elias/models/m_ik2.htm

    Public Function ftnReverseKinematics(ByVal tpcTarget As typePointCoordinate, ByRef taaArmAngles As typeArmAngles, ByRef tjcJointCoordinates As typeJointCoordinate) As Boolean
        Try
            Dim taaArmAnglesDesired As typeArmAngles = taaArmAngles
            Dim tjcJointCoordinatesDesired As typeJointCoordinate = tjcJointCoordinates
            ftnReverseKinematics = False
            frmDebug.subWriteLogs("ftnReverseKinematics: sngShoulder + 0 <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< ")
            ftnReverseKinematics = ftnFindBetterAnglesTop(tpcTarget, False, taaArmAnglesDesired, tjcJointCoordinatesDesired)
            'If ftnReverseKinematics Then subDrawArmMovement(tpcTarget, taaArmAngles, ftnConvertTo180(taaArmAngles.sngShoulder))
            If Not ftnReverseKinematics Then
                taaArmAnglesDesired = taaArmAngles
                taaArmAnglesDesired.sngShoulder = ftnConvertTo180(taaArmAngles.sngShoulder + 90)
                tjcJointCoordinatesDesired = ftnForwardKinematicsCoordinates(taaArmAnglesDesired)
                frmDebug.subWriteLogs("ftnReverseKinematics: sngShoulder + 90 <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< ")
                ftnReverseKinematics = ftnFindBetterAnglesTop(tpcTarget, False, taaArmAnglesDesired, tjcJointCoordinatesDesired)
                'If ftnReverseKinematics Then subDrawArmMovement(tpcTarget, taaArmAngles, ftnConvertTo180(taaArmAngles.sngShoulder + 90))
            End If
            If Not ftnReverseKinematics Then
                taaArmAnglesDesired = taaArmAngles
                taaArmAnglesDesired.sngShoulder = ftnConvertTo180(taaArmAngles.sngShoulder + 180)
                tjcJointCoordinatesDesired = ftnForwardKinematicsCoordinates(taaArmAnglesDesired)
                frmDebug.subWriteLogs("ftnReverseKinematics: sngShoulder + 180 <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< ")
                ftnReverseKinematics = ftnFindBetterAnglesTop(tpcTarget, False, taaArmAnglesDesired, tjcJointCoordinatesDesired)
                'If ftnReverseKinematics Then subDrawArmMovement(tpcTarget, taaArmAngles, ftnConvertTo180(taaArmAngles.sngShoulder + 180))
            End If
            If Not ftnReverseKinematics Then
                taaArmAnglesDesired = taaArmAngles
                taaArmAnglesDesired.sngShoulder = ftnConvertTo180(taaArmAngles.sngShoulder + 270)
                tjcJointCoordinatesDesired = ftnForwardKinematicsCoordinates(taaArmAnglesDesired)
                frmDebug.subWriteLogs("ftnReverseKinematics: sngShoulder + 270 <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< ")
                ftnReverseKinematics = ftnFindBetterAnglesTop(tpcTarget, False, taaArmAnglesDesired, tjcJointCoordinatesDesired)
                'If ftnReverseKinematics Then subDrawArmMovement(tpcTarget, taaArmAngles, ftnConvertTo180(taaArmAngles.sngShoulder + 270))
            End If
            If Not ftnReverseKinematics Or Not ftnValidKinematics(taaArmAnglesDesired) Then
                Beep()
            Else
                taaArmAngles = taaArmAnglesDesired
                tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAnglesDesired)
            End If

        Catch ex As Exception
            ftnReverseKinematics = False
            frmDebug.subWriteLogs(" EXCEPTION in ftnReverseKinematics " & ex.ToString)
        End Try
    End Function


    Function ftnFindBetterAnglesTop(ByVal tpcTarget As typePointCoordinate, ByVal boolDrawArm As Boolean, ByRef taaArmAngles As typeArmAngles, ByRef tjcJointCoordinates As typeJointCoordinate) As Boolean
        Dim taaStartingArmAngles As typeArmAngles = taaArmAngles
        Try
            Dim timeOut As DateTime = Now().AddSeconds(5)
            Dim tpcGripperEndOld As typePointCoordinate = tjcJointCoordinates.tpcWrist
            If booldebugArm Then
                frmDebug.subWriteLogs(">>>>> ftnFindBetterAnglesTop: ==================  START =========================================================")
                frmDebug.subWriteLogs(">>>>> ftnFindBetterAnglesTop: taaArmAngles = " & taaArmAngles.sngShoulder & ", " & taaArmAngles.sngElbow)
            End If
            While (ftnDistanceBetweenPoints(tpcTarget, tjcJointCoordinates.tpcGripperEnd) >= 10) And _
                      Not ftnPointsAreSame(tpcGripperEndOld, tjcJointCoordinates.tpcGripperEnd) And (Now() < timeOut) And Not boolShutDown
                tpcGripperEndOld = tjcJointCoordinates.tpcGripperEnd
                ftnFindBetterAngles(tpcTarget, taaArmAngles, tjcJointCoordinates) ' find incrementally better shoulder, elbow angles
                If boolDrawArm Then subDrawArm()
            End While
            ftnFindBetterAnglesTop = (ftnDistanceBetweenPoints(tpcTarget, tjcJointCoordinates.tpcGripperEnd) < 10) And ftnValidKinematics(taaArmAngles)
            If Not ftnFindBetterAnglesTop Then
                taaArmAngles = taaStartingArmAngles
                tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            End If
        Catch ex As Exception
            ftnFindBetterAnglesTop = False
            taaArmAngles = taaStartingArmAngles
            tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            frmDebug.subWriteLogs(" EXCEPTION in ftnFindBetterAnglesTop " & ex.ToString)
        End Try
    End Function

    Function ftnFindBetterAngles(ByVal tpcGripperTarget As typePointCoordinate, ByRef taaArmAngles As typeArmAngles, ByRef tjcJointCoordinates As typeJointCoordinate) As Boolean
        Dim taaStartingArmAngles As typeArmAngles = taaArmAngles
        Try
            Dim sngBestShoulderAngle As Single = taaStartingArmAngles.sngShoulder
            Dim sngBestElbowAngle As Single = taaStartingArmAngles.sngElbow
            Dim sngBestErrorDistance As Single = ftnDistanceBetweenPoints(tpcGripperTarget, tjcJointCoordinates.tpcGripperEnd)  ' error if no change in angle
            Dim sngAngleAdjustment As Single = ftnAngleAdjustment(sngBestErrorDistance)
            If booldebugArm Then frmDebug.subWriteLogs(" -- ftnFindBetterAngles: sngBestErrorDistance = " & sngBestErrorDistance & "  ///   sngAngleAdjustment = " & sngAngleAdjustment)

            ftnFindBetterAngles = ftnTryDifferentAngles(sngBestErrorDistance, tpcGripperTarget, _
                                                       ftnConvertTo180(taaStartingArmAngles.sngShoulder + sngAngleAdjustment), sngBestShoulderAngle, _
                                                        ftnConvertTo180(taaStartingArmAngles.sngElbow + sngAngleAdjustment), sngBestElbowAngle) ' + shoulder + elbow angles
            ftnFindBetterAngles = ftnFindBetterAngles Or _
                                 ftnTryDifferentAngles(sngBestErrorDistance, tpcGripperTarget, _
                                                        ftnConvertTo180(taaStartingArmAngles.sngShoulder + sngAngleAdjustment), sngBestShoulderAngle, _
                                                        ftnConvertTo180(taaStartingArmAngles.sngElbow - sngAngleAdjustment), sngBestElbowAngle) ' + shoulder - elbow angles
            ftnFindBetterAngles = ftnFindBetterAngles Or _
                                 ftnTryDifferentAngles(sngBestErrorDistance, tpcGripperTarget, _
                                                        ftnConvertTo180(taaStartingArmAngles.sngShoulder - sngAngleAdjustment), sngBestShoulderAngle, _
                                                        ftnConvertTo180(taaStartingArmAngles.sngElbow - sngAngleAdjustment), sngBestElbowAngle) ' - shoulder - elbow angles
            ftnFindBetterAngles = ftnFindBetterAngles Or _
                                 ftnTryDifferentAngles(sngBestErrorDistance, tpcGripperTarget, _
                                                        ftnConvertTo180(taaStartingArmAngles.sngShoulder - sngAngleAdjustment), sngBestShoulderAngle, _
                                                        ftnConvertTo180(taaStartingArmAngles.sngElbow + sngAngleAdjustment), sngBestElbowAngle) ' - shoulder + elbow angles
            ftnFindBetterAngles = ftnFindBetterAngles Or _
                                 ftnTryDifferentAngles(sngBestErrorDistance, tpcGripperTarget, _
                                                        ftnConvertTo180(taaStartingArmAngles.sngShoulder), sngBestShoulderAngle, _
                                                        ftnConvertTo180(taaStartingArmAngles.sngElbow + sngAngleAdjustment), sngBestElbowAngle) '  same shoulder + elbow angles
            ftnFindBetterAngles = ftnFindBetterAngles Or _
                                  ftnTryDifferentAngles(sngBestErrorDistance, tpcGripperTarget, _
                                                         ftnConvertTo180(taaStartingArmAngles.sngShoulder), sngBestShoulderAngle, _
                                                         ftnConvertTo180(taaStartingArmAngles.sngElbow - sngAngleAdjustment), sngBestElbowAngle) ' same shoulder - elbow angles
            ftnFindBetterAngles = ftnFindBetterAngles Or _
                                 ftnTryDifferentAngles(sngBestErrorDistance, tpcGripperTarget, _
                                                        ftnConvertTo180(taaStartingArmAngles.sngShoulder + sngAngleAdjustment), sngBestShoulderAngle, _
                                                        ftnConvertTo180(taaStartingArmAngles.sngElbow), sngBestElbowAngle) ' + shoulder same elbow angles
            ftnFindBetterAngles = ftnFindBetterAngles Or _
                                  ftnTryDifferentAngles(sngBestErrorDistance, tpcGripperTarget, _
                                                         ftnConvertTo180(taaStartingArmAngles.sngShoulder - sngAngleAdjustment), sngBestShoulderAngle, _
                                                         ftnConvertTo180(taaStartingArmAngles.sngElbow), sngBestElbowAngle) ' - shoulder same elbow angles
            If ftnFindBetterAngles Then
                taaArmAngles.sngShoulder = sngBestShoulderAngle
                taaArmAngles.sngElbow = sngBestElbowAngle
                taaArmAngles.sngWrist = sngBestElbowAngle   '
                taaArmAngles.sngWristRotate = sngBestElbowAngle
                tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
                If booldebugArm Then frmDebug.subWriteLogs(" ftnFindBetterAngles: BETTER taaArmAngles = " & taaArmAngles.sngShoulder & ", " & taaArmAngles.sngElbow)
            End If
            If Not ftnValidKinematics(taaArmAngles) Or Not ftnFindBetterAngles Then
                taaArmAngles = taaStartingArmAngles
                tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            End If
        Catch ex As Exception
            taaArmAngles = taaStartingArmAngles
            tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
            ftnFindBetterAngles = False
            frmDebug.subWriteLogs(" EXCEPTION in ftnFindBetterAngles " & ex.ToString)
        End Try
    End Function

    Function ftnTryDifferentAngles(ByRef sngBestErrorDistance As Single, ByVal tpcGripperTarget As typePointCoordinate, _
                                   ByVal sngShoulderAngle As Single, ByRef sngBestShoulderAngle As Single, _
                                   ByVal sngElbowAngle As Single, ByRef sngBestElbowAngle As Single) As Boolean
        Try
            ftnTryDifferentAngles = False
            Dim sngErrorDistance2 As Single
            Dim taaArmAnglesDesired As typeArmAngles = taaArmAngles
            taaArmAnglesDesired.sngShoulder = sngShoulderAngle
            taaArmAnglesDesired.sngElbow = sngElbowAngle
            taaArmAnglesDesired.sngWrist = taaArmAnglesDesired.sngElbow
            taaArmAnglesDesired.sngWristRotate = taaArmAnglesDesired.sngElbow
            Dim tjcJointCoordinatesDesired As typeJointCoordinate = ftnForwardKinematicsCoordinates(taaArmAnglesDesired)
            If booldebugArm Then
                frmDebug.subWriteLogs("    =====  ftnTryDifferentAngles: taaArmAnglesDesired >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> " & taaArmAnglesDesired.sngShoulder & ", " & taaArmAnglesDesired.sngElbow)
                frmDebug.subWriteLogs("    =====  ftnTryDifferentAngles:tpcGripperEnd =  " & tjcJointCoordinatesDesired.tpcGripperEnd.X & ", " & tjcJointCoordinatesDesired.tpcGripperEnd.Y & ", " & tjcJointCoordinatesDesired.tpcGripperEnd.Z)
            End If
            If ftnValidDesiredKinematics(taaArmAnglesDesired.sngWrist, taaArmAnglesDesired.sngShoulder) Then
                sngErrorDistance2 = ftnDistanceBetweenPoints(tpcGripperTarget, tjcJointCoordinatesDesired.tpcGripperEnd)  ' error if angle changed
                If (sngErrorDistance2 < sngBestErrorDistance) Then
                    If booldebugArm Then frmDebug.subWriteLogs("    =====  ftnTryDifferentAngles:========================= WINNER! " & sngErrorDistance2)
                    sngBestShoulderAngle = taaArmAnglesDesired.sngShoulder
                    sngBestElbowAngle = taaArmAnglesDesired.sngElbow
                    sngBestErrorDistance = sngErrorDistance2
                    ftnTryDifferentAngles = True
                Else
                    If booldebugArm Then frmDebug.subWriteLogs("    =====  ftnTryDifferentAngles:===================  " & sngErrorDistance2)
                End If
            Else
                If booldebugArm Then frmDebug.subWriteLogs("    =====  ftnTryDifferentAngles:====Not Valid ")
            End If
        Catch ex As Exception
            ftnTryDifferentAngles = False
            frmDebug.subWriteLogs(" EXCEPTION in ftnTryDifferentAngles " & ex.ToString)
        End Try
    End Function

    'Sub subDrawArmMovement(ByVal tpcTarget As typePointCoordinate, ByVal taaOriginalArmAngles As typeArmAngles, ByVal sngShoulderAngle As Single)
    '    Try
    '        taaArmAngles = ftnForwardKinematicsAngles()
    '        taaArmAngles.sngShoulder = sngShoulderAngle
    '        tjcJointCoordinates = ftnForwardKinematicsCoordinates(taaArmAngles)
    '        ftnFindBetterAnglesTop(tpcTarget, True, taaArmAngles, tjcJointCoordinates)
    '    Catch ex As Exception
    '        frmDebug.subWriteLogs(" EXCEPTION in subDrawArmMovement " & ex.ToString)
    '    End Try
    'End Sub

    Function ftnAngleAdjustment(ByVal sngError As Single) As Single
        On Error Resume Next
        If (sngError > 100) Then
            ftnAngleAdjustment = 7
        ElseIf (sngError > 30) Then
            ftnAngleAdjustment = 3
        Else
            ftnAngleAdjustment = 2
        End If
    End Function





    Function ftnValidDesiredKinematics(ByVal sngGripperAngleDesired As Single, _
                                   ByVal sngShoulderAngle As Single) As Boolean
        Try
            ftnValidDesiredKinematics = False
            Dim taaArmAnglesDesired As typeArmAngles = taaArmAngles
            taaArmAnglesDesired.sngShoulder = sngShoulderAngle
            taaArmAnglesDesired.sngElbow = sngGripperAngleDesired
            taaArmAnglesDesired.sngWrist = sngGripperAngleDesired   '
            taaArmAnglesDesired.sngWristRotate = sngGripperAngleDesired
            ftnValidDesiredKinematics = ftnValidKinematics(taaArmAnglesDesired)
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnValidDesiredKinematics " & ex.ToString)
        End Try
    End Function




    '===============================================    ================================

    Function ftnValidKinematics(ByVal taaArmAngles As typeArmAngles) As Boolean
        Try
            ftnValidKinematics = True
            Dim tjcJointCoordinates As typeJointCoordinate = ftnForwardKinematicsCoordinates(taaArmAngles)
            If (taaArmAngles.sngShoulder >= maxJointAngle) Or (taaArmAngles.sngShoulder <= -70) Then
                ftnValidKinematics = False
            ElseIf (taaArmAngles.sngElbow >= maxJointAngle) Or (taaArmAngles.sngElbow <= -maxJointAngle) Then
                ftnValidKinematics = False
            ElseIf (tjcJointCoordinates.tpcElbow.X < 0) And (tjcJointCoordinates.tpcElbow.Z < 100) Then
                ftnValidKinematics = False
            ElseIf (tjcJointCoordinates.tpcGripperEnd.X < 0) And (tjcJointCoordinates.tpcGripperEnd.Z < 300) Then
                ftnValidKinematics = False
            ElseIf (tjcJointCoordinates.tpcGripperEnd.Z < -330) Then
                ftnValidKinematics = False
            End If
            If booldebugArm Then frmDebug.subWriteLogs(" ftnValidKinematics: ftnValidKinematics = " & ftnValidKinematics)
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnValidKinematics " & ex.ToString)
        End Try
    End Function



    Function ftnPointFromPoint(ByVal tpcOrigin As typePointCoordinate, ByVal sngDistance As Single, ByVal sngAngleXZ As Single) As typePointCoordinate
        On Error Resume Next ' TODO - Add sngAngleYX
        ftnPointFromPoint.Y = tpcOrigin.Y
        ftnPointFromPoint.X = tpcOrigin.X + sngDistance * Cos(ftnDegreeToRadian(sngAngleXZ))
        ftnPointFromPoint.Z = tpcOrigin.Z + sngDistance * Sin(ftnDegreeToRadian(sngAngleXZ))
    End Function

    Function ftnDistanceBetweenPoints(ByVal tpcCoordinate1 As typePointCoordinate, ByVal tpcCoordinate2 As typePointCoordinate) As Single
        On Error Resume Next
        ftnDistanceBetweenPoints = Sqrt((tpcCoordinate2.X - tpcCoordinate1.X) ^ 2 + (tpcCoordinate2.Y - tpcCoordinate1.Y) ^ 2 + (tpcCoordinate2.Z - tpcCoordinate1.Z) ^ 2)
    End Function

    Function ftnPointsAreSame(ByVal tpcCoordinate1 As typePointCoordinate, ByVal tpcCoordinate2 As typePointCoordinate) As Single
        On Error Resume Next
        ftnPointsAreSame = (tpcCoordinate2.X = tpcCoordinate1.X) And (tpcCoordinate2.Y = tpcCoordinate1.Y) And (tpcCoordinate2.Z = tpcCoordinate1.Z)
    End Function


End Class
