﻿
'TODO - ADD watchdog, 

' FOV = 240 degrees
' 10 Hz Update Rate
' 1mm resolution
' 0.36 degrees angular resolution
' 682 Beams 

' This module makes use of only 640 beams and a  FOV of 225.22 degrees


'SCIP is a protocol used to communicate with URG. 
'Commands of SCIP protocols are used to interact with the URG. 

'SCIP commands are used to interact with URG sensor
'SCIP commands are classified as "Sensor information command", "Distance acquisition command" , and "Mode setting command".
'Setting of operation mode
'   "SCIP2.0" ... Change SCIP2.0 mode
'   "TM0", "TM1", "TM2" ... Time stamp mode
'   "SS" ... Change the baudrate
'   "BM" ... laser on.
'   "QT" ... laser off
'   "RS" ... Resets parameter, measurement suspends.
'   "CR" ... Change the rotation speed of the motor.
'Obtain status information
'   "VV" ... Acquisition of version information.
'   "PP" ... Acquisition of parameter information.
'   "II" ... Acquisition of status.
'Receive distance data
'   "GD", "GS" ... Acquisition of distance data one at a time.
'   "MD", "MS" ... Acquisition of distance data continuously.

Imports System.Threading
Imports System.IO.Ports
Imports System.Drawing
Imports System.Windows.Forms


Module modIF_URG
    Friend WithEvents comURG As New SerialPort
    Dim strMessage As String

    Public Function ftnURGInitializeTop() As Boolean
        Try
            ftnURGInitializeTop = False
            If Not boolURGConnected And Not boolShutDown Then
                strStatusDriving = "Detecting URG."
                For i = 1 To 2
                    Application.DoEvents()
                    For Each sp As String In My.Computer.Ports.SerialPortNames
                        If Not boolURGConnected And Not boolShutDown And Not ((sp = strBotComPort) And boolBotConnected) Then   ' And (strURGComPort <> "None") Then
                            frmDebug.subWriteLogs(" ftnURG_SerialConfig: Checking " & sp)
                            comURG.Close()
                            With comURG
                                .PortName = sp  'strURGComPort
                            End With
                            ftnURG_SerialOpen()
                            ftnURG_CheckConnection()
                        End If
                    Next sp
                Next i
                If boolURGConnected Then
                    frmDebug.subWriteLogs(" ftnURG_SerialConfig: Connected " & comURG.PortName)
                    strStatusDriving = "URG Connected."
                    strURGComPort = comURG.PortName
                    sngScanFOV = 225.22
                    intScanVectorsMaxRangeInches = 560 / 2.54
                    intScanVectorsMinRangeInches = 0
                    subCalculateScanBeamAngles()
                    subCalculateScanCalCoefficients()
                Else
                    strURGComPort = "None"
                End If
            Else
                boolURGConnected = ftnGetURGScanVectors()
            End If
            strStatusDriving = ""
            ftnURGInitializeTop = boolURGConnected
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnURGInitializeTop " & ex.ToString)
            boolURGConnected = False
            ftnURGInitializeTop = False
            comURG.Dispose()
        End Try
    End Function

    Public Function ftnURG_SerialOpen() As Boolean
        ' returns true if comBot is successfully opened
        Try
            If Not boolURGConnected Then
                If comURG IsNot Nothing Then
                    If Not comURG.IsOpen Then comURG.Open()
                End If
            End If
            ftnURG_SerialOpen = comURG.IsOpen
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnURG_SerialOpen " & ex.ToString)
            comURG.Dispose()
        End Try
    End Function

    Function ftnURG_CheckConnection() As Boolean
        Try
            If comURG.IsOpen And Not boolShowVectors Then
                comURG.Write("SCIP2.0" & vbLf & "RS" & vbLf & "QT" & vbLf)
                Dim ye As String = comURG.ReadExisting().Trim().Replace(vbLf, ";").Replace(";;", ";")
                Dim timeOut As DateTime = Now().AddMilliseconds(200)
                While (ye.EndsWith("00P")) And (Now() < timeOut)
                    comURG.Write("QT" & vbLf)
                    ye = comURG.ReadExisting().Trim().Replace(vbLf, ";").Replace(";;", ";")
                End While
                If (Now() < timeOut) Then
                    boolURGConnected = ftnGetURGScanVectors()
                    Dim timeGetStart As DateTime = Now()
                    boolURGConnected = ftnGetURGScanVectors()
                    intScannerDelay = Now().Subtract(timeGetStart).TotalMilliseconds
                    If boolURGConnected Then subWait(500)
                End If
            End If
            If Not boolURGConnected Then comURG.Dispose()
            ftnURG_CheckConnection = boolURGConnected
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnURG_CheckConnection " & ex.ToString)
            comURG.Dispose()
        End Try
    End Function


    Private Sub subURGportDataReceived(sender As System.Object, e As System.IO.Ports.SerialDataReceivedEventArgs) Handles comURG.DataReceived
        Try
            strMessage &= comURG.ReadExisting()
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in subURGportDataReceived " & ex.ToString)
        End Try
    End Sub

    Dim boolGetVectorsIsBusy As Boolean = False
    Public boolUrgAccessed As Boolean = False
    Public boolUrgLaserOn As Boolean = False
    Dim listData As New List(Of Integer)
    Public Function ftnGetURGScanVectors() As Boolean
        ftnGetURGScanVectors = False
        Try
            Dim timeOut As DateTime = Now().AddMilliseconds(2000)
            While boolGetVectorsIsBusy And (Now() < timeOut)
                Thread.Sleep(10)
            End While
            If (Now() >= timeOut) Then frmDebug.subWriteLogs(" ftnGetURGScanVectors: ERROR ERROR ERROR busy timeOut ")
            boolGetVectorsIsBusy = True
            If comURG.IsOpen Then
                strMessage = ""
                If Not boolUrgLaserOn Then
                    comURG.Write("BM" & vbLf) ' Laser on
                    subWait(150)
                    boolUrgLaserOn = True
                End If
                boolUrgAccessed = True
                comURG.Write("GD0044072501" & vbCrLf)
                timeOut = Now().AddMilliseconds(1000)
                Do Until ((strMessage.StartsWith("GD") And (strMessage.Length > 2133)) Or boolShutDown Or (Now() > timeOut))
                    Thread.Sleep(10)
                Loop
                If boolDebugURG Then
                    If tmiMoveInfo.boolBotIsMoving Then
                        frmDebug.subWriteLogs(ftnElapsedTimeS() & " ftnGetURGScanVectors: strMessage.Length =  " & strMessage.Length)
                    Else
                        frmDebug.subWriteLogs(ftnElapsedTimeS() & " ftnGetURGScanVectors: strMessage.Length =  " & strMessage.Length)
                    End If
                End If
                If (Now() > timeOut) Or (strMessage.Length <= 2133) Then
                    If (Now() > timeOut) And boolURGConnected Then frmDebug.subWriteLogs(" ftnGetURGScanVectors: ERROR ERROR ERROR strMessage.Length timeOut " & strMessage.Length)
                    ftnGetURGScanVectors = False
                Else
                    Dim strTemp() As String = strMessage.Split(vbLf)
                    Dim strLine As String = ""
                    listData.Clear()
                    For i As Integer = 3 To strTemp.Length - 1
                        If (strTemp(i).Length > 0) Then strLine &= strTemp(i).Substring(0, strTemp(i).Length - 1).ToString()
                    Next
                    listData.AddRange(ftnDecodeStr(strLine))
                    For intBeamNumber As Short = 0 To constMaxScannerBeams - 1  ' Convert to trvScanVectors and apply correction coefficients
                        trvScanVectors(intBeamNumber).sngRange = listData(660 - intBeamNumber)
                        trvScanVectors(intBeamNumber).sngRange = ftnCmToInches(trvScanVectors(intBeamNumber).sngRange) * sngScannerCalCoefficients(intBeamNumber)
                        If boolDebugURG Then frmDebug.subWriteVector(" ftnGetURGScanVectors: trvScanVectors(" & intBeamNumber & ") ", trvScanVectors(intBeamNumber))
                    Next intBeamNumber
                    ftnGetURGScanVectors = True
                End If
            Else
                If boolDebugURG Then frmDebug.subWriteLogs(" ftnGetURGScanVectors: comURG not open ")
            End If
            boolGetVectorsIsBusy = False
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnGetURGScanVectors " & ex.ToString)
            frmDebug.subWriteLogs("   ftnGetURGScanVectors: strMessage =  " & strMessage)
            boolGetVectorsIsBusy = False
        End Try
    End Function


    Public Sub subUrgLaserOff()
        Try
            If comURG IsNot Nothing Then
                Dim availableSerialPorts As System.Collections.ObjectModel.ReadOnlyCollection(Of String) = My.Computer.Ports.SerialPortNames
                'If boolURGConnected And Not boolShutDown And (strURGComPort <> "None") And
                '     availableSerialPorts.Contains(strURGComPort) And comURG.IsOpen Then
                If boolURGConnected And Not boolShutDown And comURG.IsOpen Then
                    comURG.Write("QT" & vbLf) ' Laser off
                End If
            End If
            boolUrgLaserOn = False
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in subUrgLaserOff " & ex.ToString)
        End Try
    End Sub

    Function ftnDecodeStr(strMessageLine As String) As List(Of Integer)
        Dim listDecode As New List(Of Integer)
        Try
            Dim strTemp As String = ""
            For i As Integer = 0 To strMessageLine.Length Step 3
                If i < strMessageLine.Length - 3 Then
                    strTemp = strMessageLine.Substring(i, 3)
                Else
                    strTemp = strMessageLine.Substring(i)
                End If
                listDecode.Add(ftnDecode(strTemp))
            Next
            If (strMessageLine.Length < (660 * 3)) Then ' ERROR- all data not received
                frmDebug.subWriteLogs(" ftnDecodeStr: ERROR ERROR ERROR Incomplete URG data received " & strMessageLine.Length)
                For i As Integer = strMessageLine.Length To 663 Step 3
                    listDecode.Add(0)
                Next
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnDecodeStr " & ex.ToString)
        End Try
        Return listDecode
    End Function

    Function ftnDecode(strSixBitEncode As String) As Integer
        Dim intDecode As Integer = 0
        Try
            For Each c As Char In strSixBitEncode
                intDecode <<= 6
                intDecode = intDecode And (Not 63)
                intDecode = intDecode Or (Asc(c) - 48)
            Next
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnDecode " & ex.ToString)
        End Try
        Return intDecode
    End Function




End Module
