﻿

' This module returns a bot pose based upon Range Scan Data.
' It is used for devices which can return a 2D range scan such as a Kinect or Laser Range scanner
Imports System.Threading
Imports System.Math
Imports System.Drawing


Module modLocalize2Dscan
    Dim bool2DScanLocalizationInProgress As Boolean = False ' dont want conflict 

    Public Function ftnLocalizeFrom2DScan_Top(ByRef tbpEstimatedPose As typeBotPose, ByVal tbpReferencePose As typeBotPose, ByVal boolBotIsMoving As Boolean, _
                                              ByRef boolGood2DLocalization As Boolean, ByRef sngRMS_Error As Single) As typeBotPose
        Try
            ftnLocalizeFrom2DScan_Top = tbpEstimatedPose
            boolGood2DLocalization = True
            If (ftnScannerEnabled() Or (boolRangersAreEnabled And Not boolBotIsMoving)) And Not boolStop And Not boolSimulatedMovement Then
                Dim tbpRawScanPose As typeBotPose
                sngRMS_Error = sngOneDotFiveSigma
                If boolBotIsMoving Then
                    intNumberOfBeams = Max(16, intNumberOfBeamsDefault \ 2) ' reduce number of vectors while moving or lost for faster localization
                Else
                    intNumberOfBeams = Max(16, intNumberOfBeamsDefault)
                End If
                boolAbortLocalization = False
                Dim strStatusDrivingRoot As String = strStatusDriving
                If bool2DScanLocalizationInProgress Then
                    Dim timeOut As DateTime = Now().AddMilliseconds(12000)
                    While bool2DScanLocalizationInProgress And Not (boolStop And Not (boolCollisionDetected Or boolCollisionPredicted)) ' wait for previous scan to finish
                        boolAbortLocalization = True
                        Thread.Sleep(25)
                        If Now() >= timeOut Then
                            ftnLocalizeFrom2DScan_Top = ftnLocalizeUsingDeadReckoning(ftnLocalizeFrom2DScan_Top, intScannerDelay, True)
                            frmDebug.subWriteLogs("  ftnLocalizeFrom2DScan_Top timeOut ")
                            Exit Function
                        End If
                    End While
                End If
                boolAbortLocalization = False
                bool2DScanLocalizationInProgress = True
                tbpRawScanPose = ftn2dLocalize(tbpEstimatedPose, boolGood2DLocalization, sngRMS_Error)
                If Not boolGood2DLocalization Then
                    ftnLocalizeFrom2DScan_Top = tbpEstimatedPose
                Else
                    If (sngRMS_Error < sngOkError) Then
                        ftnLocalizeFrom2DScan_Top = tbpRawScanPose
                    ElseIf (sngRMS_Error > sngOkError) And tmiMoveInfo.boolBotIsMoving Then
                        ftnLocalizeFrom2DScan_Top = tbpEstimatedPose
                        boolGood2DLocalization = False
                    ElseIf (sngRMS_Error < sngUnacceptableError) Then
                        ftnLocalizeFrom2DScan_Top = ftnFindAvgPose(tbpEstimatedPose, tbpRawScanPose, 0)
                    Else
                        boolGood2DLocalization = False
                        ftnLocalizeFrom2DScan_Top = tbpEstimatedPose
                    End If
                    ftnLocalizeFrom2DScan_Top = ftnMinPoseUncertainty(ftnLocalizeFrom2DScan_Top)
                    If tmiMoveInfo.boolBotIsMoving Then tbpReferencePose = tbpEstimatedPose
                    If boolGood2DLocalization Then ftnLocalizeFrom2DScan_Top = ftnValidatePose(ftnLocalizeFrom2DScan_Top, tbpReferencePose, boolGood2DLocalization, Not boolBotIsMoving, "Scan2D")
                End If
                If Not tmiMoveInfo.boolBotIsMoving And Not boolEnableRBPF Then ' update graphics and log file with results
                    strStatusDriving = strStatusDrivingRoot & " 2D=" & FormatNumber(sngRMS_Error, 2) & "; "
                    frmDebug.subWritePose(ftnElapsedTimeS() & "       Pose Derived From 2D Scan Data  ", tbpRawScanPose)
                    If boolGood2DLocalization Then
                        boolShowGreenVectors = True
                        frmDebug.subWriteLogs("                         Scan Error = " & FormatNumber(sngRMS_Error, 2) & "   (2D localization success)")
                    Else
                        frmDebug.subWriteLogs("                           Scan Error = " & FormatNumber(sngRMS_Error, 2) & "   (2D localization failed)")
                        If boolDebugLocalize2D Then
                            frmDebug.subWritePose("tbpReferencePose = ", tbpReferencePose)
                            frmDebug.subWritePose("ftnLocalizeFrom2DScan_Top = ", ftnLocalizeFrom2DScan_Top)
                            If Not ftnLocationsEqual(ftnCheckMapArrayLimits(ftnLocalizeFrom2DScan_Top.Location), ftnLocalizeFrom2DScan_Top.Location) Then
                                frmDebug.subWriteLogs(" ftnValidatePose:ftnLocationsEqual = false & vbcrlf")
                            ElseIf ftnInWall(ftnLocalizeFrom2DScan_Top.Location) Then
                                frmDebug.subWriteLogs(" ftnValidatePose:ftnInWall = false")
                            ElseIf (Not ftnLocationWithinPose(ftnLocalizeFrom2DScan_Top.Location, tbpReferencePose)) Then
                                frmDebug.subWriteLogs(" ftnValidatePose:ftnLocationWithinPose = false")
                            ElseIf Not ftnAnglesWithinUncertainty(ftnLocalizeFrom2DScan_Top.Orientation, tbpReferencePose.Orientation, tbpReferencePose.OrientationUncertainty) Then
                                frmDebug.subWriteLogs(" ftnValidatePose:ftnAnglesWithinUncertainty = false")
                            End If
                        End If
                    End If
                End If
            End If
            If boolDebugHero Or boolDebugLocalize2D Then
                Dim trvVector As typeRangeVector
                Dim tmcVectorOrigin As typeMapCoordinate = ftnLocationFromEstimatedPose(ftnLocalizeFrom2DScan_Top, tmcScanVectorOffset)
                Dim intDistanceFromWallErrorSquared As Short
                Dim boolEndPointInGreySpace As Boolean
                For intBeamNumber As Short = 0 To constMaxScannerBeams - 1 Step constMaxScannerBeams \ intNumberOfBeams
                    trvVector.sngRange = trvScanVectors(intBeamNumber).sngRange
                    If (trvVector.sngRange > 0) And (trvVector.sngRange < intScanVectorsMaxRangeInches) Then  ' 
                        trvVector.sngAngle = ftnConvertTo360(trvScanVectors(intBeamNumber).sngAngle + ftnLocalizeFrom2DScan_Top.Orientation)
                        intDistanceFromWallErrorSquared = ftnDistanceFromWallErrorSquared(tmcVectorOrigin, trvVector, False, boolEndPointInGreySpace)
                        frmDebug.subWriteLogs(" Beam " & intBeamNumber & ":  Range = " & trvVector.sngRange & ":  Error = " & intDistanceFromWallErrorSquared)
                    End If
                Next intBeamNumber
            End If
            bool2DScanLocalizationInProgress = False
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnLocalizeFrom2DScan_Top :" & ex.ToString)
            ftnLocalizeFrom2DScan_Top = tbpBotPose
            sngRMS_Error = sngOneDotFiveSigma
        End Try
    End Function




    Function ftn2dLocalize(ByRef tbpEstimatedPose As typeBotPose,
                           ByRef boolGood2DLocalization As Boolean, ByRef sngRMS_Error As Single) As typeBotPose
        ftn2dLocalize = tbpEstimatedPose
        Try
            sngRMS_Error = sngOneDotFiveSigma
            boolGood2DLocalization = False
            If ftnGetScanVectors() Then
                boolShowVectors = Not tmiMoveInfo.boolBotIsMoving
                If ftnNonFlatWall() Then
                    ftn2dLocalize = ftnPoseFrom2DScan(tbpEstimatedPose, sngRMS_Error) '
                    boolGood2DLocalization = sngRMS_Error < sngUnacceptableError
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftn2dLocalize :" & ex.ToString)
            ftn2dLocalize = tbpEstimatedPose
            sngRMS_Error = sngOneDotFiveSigma
        End Try
    End Function

 


    Public Function ftnPoseFrom2DScan(ByRef tbpEstimatedPose As typeBotPose, ByRef sngRMS_Error As Single) As typeBotPose
        Try
            ftnPoseFrom2DScan = tbpEstimatedPose
            If tmiMoveInfo.boolBotIsMoving Then
                If tmiMoveInfo.tomTypeOfMove = TypeOfMove.TURN Then
                    ftnPoseFrom2DScan.OrientationUncertainty = 20
                Else
                    ftnPoseFrom2DScan.OrientationUncertainty = 8
                End If
                ftnPoseFrom2DScan.LocationUncertainty.X = 8
                ftnPoseFrom2DScan.LocationUncertainty.Y = 8
            End If
            ftnPoseFrom2DScan = ftnPoseWithBestFit(ftnPoseFrom2DScan, sngRMS_Error, 4, 1)
            ftnPoseFrom2DScan.LocationUncertainty.X = 3
            ftnPoseFrom2DScan.LocationUncertainty.Y = 3
            ftnPoseFrom2DScan.OrientationUncertainty = 3
            If (sngRMS_Error < sngOneDotFiveSigma) Then _
                ftnPoseFrom2DScan = ftnPoseWithBestFit(ftnPoseFrom2DScan, sngRMS_Error, 1, 1)
            ftnPoseFrom2DScan = ftnPoseUncertainty(tbpEstimatedPose, ftnPoseFrom2DScan, sngRMS_Error, (sngRMS_Error < sngUnacceptableError))
        Catch ex As Exception
            ftnPoseFrom2DScan = tbpEstimatedPose
            sngRMS_Error = sngOneDotFiveSigma
            frmDebug.subWriteLogs(" EXCEPTION in ftnPoseFrom2DScan " & ex.ToString)
        End Try
    End Function

    Function ftnPoseUncertainty(ByRef tbpEstimatedPose As typeBotPose, ByVal tbpLocalizedPose As typeBotPose, ByRef sngRMS_Error As Single, ByRef boolValidLocalization As Boolean) As typeBotPose
        Try
            If Not boolValidLocalization Then
                ftnPoseUncertainty = tbpEstimatedPose
            Else
                ftnPoseUncertainty = tbpLocalizedPose
                If (sngRMS_Error < sngLowError) Then
                    ftnPoseUncertainty.OrientationUncertainty = 10
                    ftnPoseUncertainty.LocationUncertainty.X = 10
                    ftnPoseUncertainty.LocationUncertainty.Y = 10
                ElseIf (sngRMS_Error < sngOkError) Then
                    ftnPoseUncertainty.OrientationUncertainty = 15
                    ftnPoseUncertainty.LocationUncertainty.X = 10
                    ftnPoseUncertainty.LocationUncertainty.Y = 10
                Else
                    ftnPoseUncertainty.OrientationUncertainty = 15
                    ftnPoseUncertainty.LocationUncertainty.X = 15
                    ftnPoseUncertainty.LocationUncertainty.Y = 15
                End If
            End If
        Catch ex As Exception
            ftnPoseUncertainty = tbpEstimatedPose
            frmDebug.subWriteLogs(" EXCEPTION in ftnPoseUncertainty " & ex.ToString)
        End Try
    End Function


    Dim boolAbortLocalization As Boolean = False
    Function ftnPoseWithBestFit(ByRef tbpEstimatedPose As typeBotPose, ByRef sngRmsErrorBest As Single,
                                                 ByVal intLocationStep As Short, ByVal intOrientationStep As Short) As typeBotPose
        Try ' Find the pose within the bot's uncertainty region that most closely resembles the range data
            ftnPoseWithBestFit = tbpEstimatedPose
            sngRmsErrorBest = sngOneDotFiveSigma
            If boolStop And Not (boolCollisionDetected Or boolCollisionPredicted) Or boolAbortLocalization Then Exit Function
            Dim intOrientationStart, intOrientationEnd As Short
            Dim tbpPossiblePose As typeBotPose = tbpEstimatedPose
            Dim intStartX, intEndX, intStartY, intEndY As Short
            Dim sngRmsError As Single
            intStartX = tbpEstimatedPose.Location.X - tbpEstimatedPose.LocationUncertainty.X
            intEndX = tbpEstimatedPose.Location.X + tbpEstimatedPose.LocationUncertainty.X
            intStartY = tbpEstimatedPose.Location.Y - tbpEstimatedPose.LocationUncertainty.Y
            intEndY = tbpEstimatedPose.Location.Y + tbpEstimatedPose.LocationUncertainty.Y
            subCheckMapArrayLimits(intStartX, intEndX, intStartY, intEndY)
            subFindStartEndAngles(tbpEstimatedPose, intOrientationStart, intOrientationEnd)
            If boolDebugRBPF Then
                frmDebug.subWriteLogs(" ftnPoseWithBestFit:  intStartX, intEndX " & intStartX & " to " & intEndX)
                frmDebug.subWriteLogs(" ftnPoseWithBestFit:  intStartY, intEndY " & intStartY & " to " & intEndY)
                frmDebug.subWriteLogs(" ftnPoseWithBestFit:  Start, Stop " & intOrientationStart & " to " & intOrientationEnd)
            End If
            If Not ftnInWall_NoChk(tbpPossiblePose.Location) Then _
                sngRmsErrorBest = ftnScanToMapRMS_Error(tbpEstimatedPose, False) ' Find  error at estimated pose
            'Dim sngDebugRms As Single = sngRmsErrorBest
            'Dim strLine As String = ""
            For tbpPossiblePose.Location.Y = intStartY To intEndY Step intLocationStep
                'strLine = intStartX & ", " & tbpPossiblePose.Location.Y & ": "
                For tbpPossiblePose.Location.X = intStartX To intEndX Step intLocationStep ' try to find a lower error pose 
                    If Not ftnInWall_NoChk(tbpPossiblePose.Location) Then
                        'sngDebugRms = 6
                        For intOrientation As Short = intOrientationStart To intOrientationEnd Step intOrientationStep
                            tbpPossiblePose.Orientation = ftnConvertTo360(intOrientation)
                            sngRmsError = ftnScanToMapRMS_Error(tbpPossiblePose, True) ' Find End Point Model error at this pose
                            'If (sngRmsError < sngDebugRms) Then sngDebugRms = sngRmsError
                            If (sngRmsError < sngRmsErrorBest) Then ' check for intervening walls
                                sngRmsError = ftnScanToMapRMS_Error(tbpPossiblePose, False) ' Find Ray Cast Model error at this pose
                                'If (sngRmsError < sngDebugRms) Then sngDebugRms = sngRmsError
                                If (sngRmsError < sngRmsErrorBest) Then ' 
                                    sngRmsErrorBest = sngRmsError
                                    ftnPoseWithBestFit = tbpPossiblePose
                                End If
                            End If
                        Next intOrientation
                        '    strLine &= FormatNumber(sngDebugRms, 2) & " "
                        'Else
                        '    strLine &= "---- "
                    End If
                Next tbpPossiblePose.Location.X
                'If boolDebugRBPF Then
                '    frmDebug.subWriteLogs("ftnPoseWithBestFit: " & strLine)
                '    Windows.Forms.Application.DoEvents()
                'End If
            Next tbpPossiblePose.Location.Y
        Catch ex As Exception
            ftnPoseWithBestFit = tbpEstimatedPose
            sngRmsErrorBest = sngOneDotFiveSigma
            frmDebug.subWriteLogs(" EXCEPTION in ftnPoseWithBestFit " & ex.ToString)
        End Try
    End Function

     Function ftnScanToMapRMS_Error(ByRef tbpPossiblePose As typeBotPose, ByRef boolEndPointModel As Boolean) As Single
        ' Find the correlation of a particular pose to the range data 
        ' boolEndPointModel = Don't check for intervening walls
        Dim intDistanceFromWallErrorSquared As Short
        Dim intValidBeamCounter As Short = 0
        Dim intDistanceFromWallErrorSquaredSum As Short = 0
        Dim boolEndPointInGreySpace As Boolean = False
        Try
            ftnScanToMapRMS_Error = 0
            Dim trvVector As typeRangeVector
            Dim tmcVectorOrigin As typeMapCoordinate = ftnLocationFromEstimatedPose(tbpPossiblePose, tmcScanVectorOffset)
            For intBeamNumber As Short = 0 To constMaxScannerBeams - 1 Step constMaxScannerBeams \ intNumberOfBeams
                trvVector.sngRange = trvScanVectors(intBeamNumber).sngRange
                If (trvVector.sngRange > 0) And (trvVector.sngRange < intScanVectorsMaxRangeInches) Then  ' 
                    trvVector.sngAngle = ftnConvertTo360(trvScanVectors(intBeamNumber).sngAngle + tbpPossiblePose.Orientation)
                    intDistanceFromWallErrorSquared = ftnDistanceFromWallErrorSquared(tmcVectorOrigin, trvVector, boolEndPointModel, boolEndPointInGreySpace)
                    If (intDistanceFromWallErrorSquared < sngOneSigmaSquared) Then
                        intValidBeamCounter += 3
                    ElseIf (intDistanceFromWallErrorSquared < intOneDotFiveSigmaSquared) Then
                        intValidBeamCounter += 1
                    ElseIf boolEndPointInGreySpace Then
                        intValidBeamCounter += 1 ' Grey zone = cluttered environment / may have false readings
                    End If
                    intDistanceFromWallErrorSquaredSum += intDistanceFromWallErrorSquared
                End If
            Next intBeamNumber
            intValidBeamCounter = intValidBeamCounter \ 3
            If False And boolSlam Then
                ftnScanToMapRMS_Error = Sqrt(intDistanceFromWallErrorSquaredSum / intNumberOfBeams)
                If boolURGConnected Then
                    ftnScanToMapRMS_Error *= 0.7  ' wider FOV then Kinects yields poorer fit, therefore higher error??
                ElseIf boolHeroSelected And Not boolKinectConnected Then
                    ftnScanToMapRMS_Error *= 0.5
                End If
                'If boolEnableRBPF Then ftnScanToMapRMS_Error *= 0.5 ' reduced range of intDistanceFromWallSquaredArray ??
                ftnScanToMapRMS_Error = Min(ftnScanToMapRMS_Error, sngOneDotFiveSigma)
                ftnScanToMapRMS_Error = Max(ftnScanToMapRMS_Error, 0.25)
            ElseIf (intValidBeamCounter > intNumberOfBeams * 0.4) Or boolHeroSelected And (intValidBeamCounter > intNumberOfBeams * 0.4) Then
                ftnScanToMapRMS_Error = Sqrt(intDistanceFromWallErrorSquaredSum / intValidBeamCounter)
                If boolURGConnected Then
                    ftnScanToMapRMS_Error *= 0.7  ' wider FOV then Kinects yields poorer fit, therefore higher error??
                ElseIf boolHeroSelected And Not boolKinectConnected Then
                    ftnScanToMapRMS_Error *= 0.5
                End If
                If boolEnableRBPF Then ftnScanToMapRMS_Error *= 0.5 ' reduced range of intDistanceFromWallSquaredArray ??
                ftnScanToMapRMS_Error = Min(ftnScanToMapRMS_Error, sngOneDotFiveSigma)
                ftnScanToMapRMS_Error = Max(ftnScanToMapRMS_Error, 0.25)
            Else
                ftnScanToMapRMS_Error = sngOneDotFiveSigma
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXECEPTION in ftnScanToMapRMS_Error " & ex.ToString)
            ftnScanToMapRMS_Error = sngOneDotFiveSigma
        End Try
    End Function

 
  
    '--------------------------------------  Non Flat Wall  ----------------------------------------------------------------------

    Public Function ftnNonFlatWall() As Boolean
        ' long flat walls which fill the Scanner's FOV can cause localization errors
        ftnNonFlatWall = False
        Try
            If boolURGConnected Then
                ftnNonFlatWall = True
            Else
                Dim sngScannerRange(constMaxScannerBeams) As Single
                Dim sngScannerSlope(constMaxScannerBeams) As Single
                Dim intBeamInterval As Short = constMaxScannerBeams \ intNumberOfBeams
                Dim intTwoSamplesPerVector As Short = intBeamInterval + intBeamInterval
                For intBeamNumber As Short = 0 To constMaxScannerBeams - 1 Step intBeamInterval
                    sngScannerRange(intBeamNumber) = trvScanVectors(intBeamNumber).sngRange
                Next intBeamNumber
                If (sngScannerRange(intTwoSamplesPerVector) = 0) Then sngScannerRange(intTwoSamplesPerVector) = sngScannerRange(3 * intBeamInterval)
                If (sngScannerRange(intBeamInterval) = 0) Then sngScannerRange(intBeamInterval) = sngScannerRange(intTwoSamplesPerVector)
                If (sngScannerRange(0) = 0) Then sngScannerRange(0) = sngScannerRange(intBeamInterval)
                sngScannerSlope(intBeamInterval) = sngScannerRange(intBeamInterval) - sngScannerRange(0)
                Dim intI_MinusSamplesPerVector As Short
                For I As Short = intTwoSamplesPerVector To constMaxScannerBeams - 1 Step intBeamInterval
                    intI_MinusSamplesPerVector = I - intBeamInterval
                    If (sngScannerRange(I) = 0) Then sngScannerRange(I) = sngScannerRange(intI_MinusSamplesPerVector) ' get rid of range = 0 entries
                    sngScannerSlope(I) = sngScannerRange(I) - sngScannerRange(intI_MinusSamplesPerVector) ' calculate slopes
                    If (sngScannerRange(I) <> 0) And (sngScannerRange(intI_MinusSamplesPerVector) <> 0) Then
                        If (Abs(sngScannerRange(I) - sngScannerRange(intI_MinusSamplesPerVector)) > 3) Then ' Exit if Spike (Discontinuity)
                            ftnNonFlatWall = True
                            Exit Function
                        End If
                    End If
                Next I
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnNonFlatWall " & ex.ToString)
        End Try
    End Function




End Module
