﻿Imports System.Math
Imports System.Drawing
Imports System.Windows.Forms

' TODO investigate LockBits  https://msdn.microsoft.com/en-us/library/5ey6h79d(v=vs.110).aspx 

Module modMapArray

    ''''''''''''''''''''''''''''''''''''''''''''''
    '
    '     Map Array Module: 
    '
    '  This module is responsible for:
    '    - converting a map (bmpMap) to a byte array (byteNavigationArray)
    '              0 = free,  255 = occupied
    '    - adding a buffer space around occupied cells 
    '
    ''''''''''''''''''''''''''''''''''''''''''''''
    Public Sub subConvertMapToArray()
        ' Occupied = (byteMapGreenPixel = 0)  
        Try
            If boolDebugRBPF Then frmDebug.subWriteLogs(ftnElapsedTimeS() & " subConvertMapToArray Start  ")
            strStatusDriving = "Generating Map Array"
            frmDebug.subWriteLogs(" Generating Map Array")
            ReDim byteNavigationArray(imgMap.Width, imgMap.Height)
            ReDim boolOccupiedCell(imgMap.Width, imgMap.Height)
            ReDim intDistanceFromWallSquaredArray(imgMap.Width, imgMap.Height)
            If Not boolSlam Then
                Dim byteMapGreenPixel As Byte
                ReDim boolGreySpace(imgMap.Width, imgMap.Height)
                ReDim boolNoGo(imgMap.Width, imgMap.Height)
                ReDim boolCompassIsSuspect(imgMap.Width, imgMap.Height)
                For intX As Short = 0 To imgMap.Width - 1   ' Convert the map image into byteNavigationArray.   0 = free,  255 = occupied
                    For intY As Short = 0 To imgMap.Height - 1
                        byteMapGreenPixel = bmpMap.GetPixel(intX, intY).G
                        If (byteMapGreenPixel = 0) Then
                            boolOccupiedCell(intX, intY) = True
                            byteNavigationArray(intX, intY) = 255
                        Else
                            boolGreySpace(intX, intY) = (byteMapGreenPixel = 223)
                            boolNoGo(intX, intY) = (byteMapGreenPixel = 1)
                            boolCompassIsSuspect(intX, intY) = (byteMapGreenPixel = 216)
                        End If
                    Next intY
                Next intX
            Else
                For intX As Short = 0 To imgMap.Width - 1   ' Convert the map image into byteNavigationArray.   0 = free,  255 = occupied
                    For intY As Short = 0 To imgMap.Height - 1
                        If (bmpMap.GetPixel(intX, intY).G = 0) Then
                            boolOccupiedCell(intX, intY) = True
                            byteNavigationArray(intX, intY) = 255
                        End If
                    Next intY
                Next intX
            End If
            subGenerateDistanceFromWallSquaredArray()
            strStatusDriving = "Map Array Generated "
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in subConvertMapToArray " & ex.ToString)
        End Try
    End Sub

    Public Sub subGenerateDistanceFromWallSquaredArray() ' TODO Set periphery = intOneDotFiveSigmaSquared.  Speeds up ftnDistanceFromWallErrorSquared as dont have to check for ftnLocationOnMapEdge
        Dim intX_Min, intX_Max, intY_Min, intY_Max As Short
        Dim intDistanceFromWallSquared As Short
        Dim tmcLocation1, tmcLocation2 As typeMapCoordinate
        Dim intMaxPathDistanceFromWall As Short = Sqrt(intMaxPathDistanceFromWallSquared)
        Dim intX_Diff, intY_Diff, intX, intY As Short
        Try
            For tmcLocation1.X = 1 To imgMap.Width - 2 ' Mark wall edges as Distance = 0
                For tmcLocation1.Y = 1 To imgMap.Height - 2
                    If boolOccupiedCell(tmcLocation1.X, tmcLocation1.Y) Then
                        If (Not boolOccupiedCell(tmcLocation1.X, tmcLocation1.Y + 1)) Then
                            intDistanceFromWallSquaredArray(tmcLocation1.X, tmcLocation1.Y) = 0
                        ElseIf (Not boolOccupiedCell(tmcLocation1.X, tmcLocation1.Y - 1)) Then
                            intDistanceFromWallSquaredArray(tmcLocation1.X, tmcLocation1.Y) = 0
                        ElseIf (Not boolOccupiedCell(tmcLocation1.X + 1, tmcLocation1.Y)) Then
                            intDistanceFromWallSquaredArray(tmcLocation1.X, tmcLocation1.Y) = 0
                        ElseIf (Not boolOccupiedCell(tmcLocation1.X - 1, tmcLocation1.Y)) Then
                            intDistanceFromWallSquaredArray(tmcLocation1.X, tmcLocation1.Y) = 0
                        Else
                            intDistanceFromWallSquaredArray(tmcLocation1.X, tmcLocation1.Y) = intMaxPathDistanceFromWallSquared
                        End If
                    Else
                        intDistanceFromWallSquaredArray(tmcLocation1.X, tmcLocation1.Y) = intMaxPathDistanceFromWallSquared
                    End If
                Next tmcLocation1.Y
            Next tmcLocation1.X

            For intX = 0 To imgMap.Width - 1
                For intY = 0 To imgMap.Height - 1
                    If (intDistanceFromWallSquaredArray(intX, intY) = 0) Then  ' location is a wall edge.  Populate nearby cells with distance from this location.
                        intX_Min = (intX - intMaxPathDistanceFromWall)
                        intX_Max = (intX + intMaxPathDistanceFromWall)
                        intY_Min = (intY - intMaxPathDistanceFromWall)
                        intY_Max = (intY + intMaxPathDistanceFromWall)
                        subCheckMapArrayLimits(intX_Min, intX_Max, intY_Min, intY_Max)
                        For tmcLocation2.X = intX_Min To intX_Max ' left and right neighbors
                            intX_Diff = Abs(tmcLocation2.X - intX)
                            For tmcLocation2.Y = intY_Min To intY_Max ' up and down neighbors
                                intY_Diff = Abs(tmcLocation2.Y - intY)
                                intDistanceFromWallSquared = intSquareDistanceArray(intX_Diff, intY_Diff)
                                intDistanceFromWallSquaredArray(tmcLocation2.X, tmcLocation2.Y) = Min(intDistanceFromWallSquaredArray(tmcLocation2.X, tmcLocation2.Y), intDistanceFromWallSquared)
                            Next tmcLocation2.Y
                        Next tmcLocation2.X
                    End If
                Next intY
            Next intX
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in subGenerateDistanceFromWallSquaredArray " & ex.Message)
            frmDebug.subWriteLogs("    tmcLocation1:  " & tmcLocation1.X & ", " & tmcLocation1.Y)
            frmDebug.subWriteLogs("    tmcLocation2:  " & tmcLocation2.X & ", " & tmcLocation2.Y)
            frmDebug.subWriteLogs("    intX_Min:  " & intX_Min)
            frmDebug.subWriteLogs("    intX_Max:  " & intX_Max)
            frmDebug.subWriteLogs("    intY_Min:  " & intY_Min)
            frmDebug.subWriteLogs("    intY_Max:  " & intY_Max)
            frmDebug.subWriteLogs("    intX:  " & intY_Max)
            frmDebug.subWriteLogs("    intY:  " & intY_Max)
        End Try
    End Sub



End Module
